/*
 * Copyright (c) 2001-2005 The Regents of The University of Michigan
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met: redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer;
 * redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution;
 * neither the name of the copyright holders nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Steve Reinhardt
 *          Lisa Hsu
 *          Nathan Binkert
 *          Steve Raasch
 */

#include <iomanip>

#include "base/loader/symtab.hh"
#include "cpu/base.hh"
#include "cpu/exetrace.hh"
#include "cpu/static_inst.hh"
#include "cpu/thread_context.hh"
#include "enums/OpClass.hh"

using namespace std;
using namespace TheISA;

namespace Trace {

#include "ftr.hh"
#define NOT_A_REGISTER 99

static inline uint8_t integerRegOnly ( int x )
{
#if THE_ISA == ARM_ISA
    const int rmax = 14 ;
#else
    const int rmax = 31 ;
#endif

    if (( x >= 0 ) && ( x <= rmax ))
    {
        return (uint8_t) x ;
    } else {
        return NOT_A_REGISTER ;
    }
}


void 
traceSysCall(int callnum, unsigned retval, ThreadContext *tc)
{
    if (IsOn(ExecFTR)) {
        ostream &outs = Trace::output();

        FastTraceRecord ftr ;

        memset ( & ftr , 0 , sizeof ( FastTraceRecord ) ) ;

        ftr.sync = '\n';
        ftr.out_reg_number = 0 ;
        ftr.optype = 'T' ;
        ftr.out_value = retval;
        ftr.ea_value = callnum;
        ftr.cost = 1 ;
        outs.write((const char *) & ftr, sizeof(FastTraceRecord));
    }
}

void
ExeTracerRecord::dumpTicks(ostream &outs)
{
    ccprintf(outs, "%7d: ", when);
}

void
Trace::ExeTracerRecord::traceInst(StaticInstPtr inst, bool ran)
{
    ostream &outs = Trace::output();

    if (IsOn(ExecTicks))
        dumpTicks(outs);

    outs << thread->getCpuPtr()->name() << " ";

    if (IsOn(ExecSpeculative))
        outs << (misspeculating ? "-" : "+") << " ";

    if (IsOn(ExecThread))
        outs << "T" << thread->threadId() << " : ";

    std::string sym_str;
    Addr sym_addr;
    if (debugSymbolTable
        && IsOn(ExecSymbol)
#if FULL_SYSTEM
        && !inUserMode(thread)
#endif
        && debugSymbolTable->findNearestSymbol(PC, sym_str, sym_addr)) {
        if (PC != sym_addr)
            sym_str += csprintf("+%d", PC - sym_addr);
        outs << "@" << sym_str;
    }
    else {
        outs << "0x" << hex << PC;
    }

    if (inst->isMicroop()) {
        outs << "." << setw(2) << dec << upc;
    } else {
        outs << "   ";
    }

    outs << " : ";

    //
    //  Print decoded instruction
    //

    outs << setw(26) << left;
    outs << inst->disassemble(PC, debugSymbolTable);

    if (ran) {
        outs << " : ";

        if (IsOn(ExecOpClass)) {
            outs << Enums::OpClassStrings[inst->opClass()] << " : ";
        }

        if (IsOn(ExecResult) && data_status != DataInvalid) {
            ccprintf(outs, " D=%#018x", data.as_int);
        }

        if (IsOn(ExecEffAddr) && addr_valid)
            outs << " A=0x" << hex << addr;

        if (IsOn(ExecFetchSeq) && fetch_seq_valid)
            outs << "  FetchSeq=" << dec << fetch_seq;

        if (IsOn(ExecCPSeq) && cp_seq_valid)
            outs << "  CPSeq=" << dec << cp_seq;

        if (inst->isControl()) {
            ccprintf(outs, " Branch=[");
            if (inst->isCondCtrl()) {
                ccprintf(outs, "C");
            }
            if (inst->isSerializeAfter()) {
                ccprintf(outs, "S");
            }
            ccprintf(outs, "]");
        }
    }

    //
    //  End of line...
    //
    outs << endl;
}

void
Trace::ExeTracerRecord::dump()
{
    if (IsOn(ExecFTR)) {
        ostream &outs = Trace::output();
        FastTraceRecord ftr ;
        //std::string dis = 
        //            staticInst->disassemble(PC, debugSymbolTable);
        bool is_store_0 = false;
        int offset = 0 ;

        ftr.sync = '\n';
        ftr.cost = 1 ;

        if (staticInst->isControl()) {
            if (staticInst->isSerializeAfter()) {
                // Serialising operation
                ftr.optype = 'X' ;
            } else if (staticInst->isCondCtrl()) {
                // Branch - conditional control
                ftr.optype = 'B' ;
            } else {
                // Jump - unconditional control (destination may be unknown)
                ftr.optype = 'J' ;
            }
        } else if (staticInst->isLoad()) {
            ftr.optype = 'L';
            ftr.cost = 2;
        } else if (staticInst->isStore()) {
            ftr.optype = 'S';
            ftr.cost = 2;
            is_store_0 = true;
            offset ++ ;
        } else if (staticInst->isInteger()) {
            ftr.optype = 'A';
            if (staticInst->opClass() == Enums::IntMult) {
                ftr.cost = 3;
            } else if (staticInst->opClass() != Enums::IntAlu) {
                ftr.optype = '?';
            }
        } else {
            ftr.optype = 'O';
        }

        ftr.out_reg_number = NOT_A_REGISTER;
        ftr.in0_reg_number = NOT_A_REGISTER;
        ftr.in1_reg_number = NOT_A_REGISTER;
        ftr.store_reg_number = NOT_A_REGISTER;
        ftr.out_value = ~0 ;
        if (staticInst->numDestRegs() > 0) {
            ftr.out_reg_number = integerRegOnly ( 
                                staticInst->destRegIdx(0) ) ;
#if THE_ISA == ARM_ISA
            if ( ftr.out_reg_number != NOT_A_REGISTER ) {
                ftr.out_value = thread->readIntReg(
                        ftr.out_reg_number);
            }
#endif
        }
        if (staticInst->numSrcRegs() > offset) {
            ftr.in0_reg_number = integerRegOnly ( 
                                staticInst->srcRegIdx(offset) ) ;
            offset ++ ;
            if (staticInst->numSrcRegs() > offset) {
                ftr.in1_reg_number = integerRegOnly ( 
                                staticInst->srcRegIdx(offset) ) ;
            }
        }
        if (is_store_0) {
            ftr.store_reg_number = integerRegOnly (
                            staticInst->srcRegIdx(0));
        }
        if (data_status == DataInvalid) {
            ftr.out_value = ~0 ;
        } else {
#if THE_ISA == ALPHA_ISA
            // Not reliable on ARM, since some operations update more
            // than one register and this causes confusion.
            ftr.out_value = data.as_int;
#endif
        }
        ftr.ea_value = addr_valid ? addr : ( ~0 );
        ftr.pc_value = PC;
        ftr.when = when;
        outs.write((const char *) & ftr, sizeof(FastTraceRecord));
        return;
    }
    /*
     * The behavior this check tries to achieve is that if ExecMacro is on,
     * the macroop will be printed. If it's on and microops are also on, it's
     * printed before the microops start printing to give context. If the
     * microops aren't printed, then it's printed only when the final microop
     * finishes. Macroops then behave like regular instructions and don't
     * complete/print when they fault.
     */
    if (IsOn(ExecMacro) && staticInst->isMicroop() &&
            ((IsOn(ExecMicro) &&
             macroStaticInst && staticInst->isFirstMicroop()) ||
            (!IsOn(ExecMicro) &&
             macroStaticInst && staticInst->isLastMicroop()))) {
        traceInst(macroStaticInst, false);
    }
    if (IsOn(ExecMicro) || !staticInst->isMicroop()) {
        traceInst(staticInst, true);
    }
}

/* namespace Trace */ }

////////////////////////////////////////////////////////////////////////
//
//  ExeTracer Simulation Object
//
Trace::ExeTracer *
ExeTracerParams::create()
{
    return new Trace::ExeTracer(this);
};
